 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
import json
from redfish import LegacyRestClient
from get_resource_directory import get_resource_directory

def get_LogicalDrives(restobj):
    resource_instances = get_resource_directory(restobj)
    if resource_instances:
        #Get URI from resource directory
        for instance in resource_instances:
            if "Collection.1.0.0" in instance.Type:
                if "HpSmartStorageArrayController." in instance.MemberType:
                    array_controllers = instance.href
                    break
    collection = restobj.get(array_controllers)
    for instance in collection.obj.links.Member:
        controller = restobj.get(instance.href)
        sys.stdout.write("Location: %s\n" % controller.obj.Location)
        logicaldrivecollection = restobj.get(controller.obj.links.LogicalDrives.href)
        for logicaldrive in logicaldrivecollection.obj.links.Member:
            drive = restobj.get(logicaldrive.href)
            sys.stdout.write("\t%s\n" % json.dumps(drive.dict))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # SYSTEM_URL = None
    # LOGIN_ACCOUNT = None
    # LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT =  "admin"
    LOGIN_PASSWORD =  "password"

    # Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    sys.stdout.write("\nEXAMPLE 44: Dump LogicalDrive details\n")
    REST_OBJ.login()
    get_LogicalDrives(REST_OBJ)
    REST_OBJ.logout()



###
# Copyright 2019 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###
# -*- coding: utf-8 -*-
"""A threaded version of RIS _load for quicker searching"""
#---------Imports---------

import logging
import threading

#Added for py3 compatibility
import six

from six.moves.queue import Empty
from six.moves.urllib.parse import urlparse, urlunparse

import jsonpath_rw

import redfish.ris

#---------End of imports---------

#---------Debug logger---------

LOGGER = logging.getLogger(__name__)

#---------End of debug logger---------

class LoadWorker(threading.Thread):
    """A threaded implementation of _load for quicker crawling""" 
    def __init__(self, queue):
        threading.Thread.__init__(self)
        self.queue = queue
        self.exception = None

    def run(self):
        """Main worker function"""
        try:
            while True:
                (path, includelogs, loadcomplete, crawl, rel,\
                 init, prevpath, originaluri, theobj) = self.queue.get()
                if path == includelogs == loadcomplete == crawl == rel == init == \
                    prevpath == originaluri == theobj =='KILL':
                    break

                if path.endswith("?page=1") and not loadcomplete:
                    #Don't download schemas in crawl unless we are loading absolutely everything
                    self.queue.task_done()
                    continue
                elif not includelogs and crawl:
                    #Only include logs when asked as there can be an extreme amount of entries
                    if "/log" in path.lower():
                        self.queue.task_done()
                        continue

                #TODO: need to find a better way to support non ascii characters
                path = path.replace("|", "%7C")
                #remove fragments
                newpath = urlparse(path)
                newpath = list(newpath[:])
                newpath[-1] = ''
                path = urlunparse(tuple(newpath))

                if prevpath and prevpath != path:
                    theobj.ctree[prevpath].update([path])
                if not rel:
                    if path.lower() in theobj.visited_urls:
                        self.queue.task_done()
                        continue
                LOGGER.debug('_loading %s', path)

                resp = theobj._client().get(path)

                if resp.status != 200 and path.lower() == theobj.typepath.defs.biospath:
                    self.queue.task_done()
                    raise redfish.ris.ris.BiosUnregisteredError()
                elif resp.status == 401:
                    self.queue.task_done()
                    raise redfish.ris.ris.SessionExpired("Invalid session. Please logout and "\
                                            "log back in or include credentials.")
                elif resp.status not in (201, 200):
                    theobj.removepath(path)
                    self.queue.task_done()
                    continue

                theobj.update_member(resp=resp, path=path, init=init)

                fpath = lambda pa, path: path if pa.endswith(theobj.typepath.defs.hrefstring) and \
                    pa.startswith((theobj.collstr, 'Entries')) else None

                #follow all the href attributes
                if theobj.is_redfish:
                    jsonpath_expr = jsonpath_rw.parse("$..'@odata.id'")
                else:
                    jsonpath_expr = jsonpath_rw.parse('$..href')
                matches = jsonpath_expr.find(resp.dict)

                if 'links' in resp.dict and 'NextPage' in resp.dict['links']:
                    if originaluri:
                        next_link_uri = originaluri + '?page=' + \
                                        str(resp.dict['links']['NextPage']['page'])
                        href = '%s' % next_link_uri
    
                        theobj.get_queue.put((href, includelogs, loadcomplete, crawl, \
                                   rel, init, None, originaluri, theobj))
                    else:
                        next_link_uri = path + '?page=' + \
                                        str(resp.dict['links']['NextPage']['page'])
    
                        href = '%s' % next_link_uri
                        theobj.get_queue.put((href, includelogs, loadcomplete, crawl, \
                                   rel, init, None, path, theobj))

                #Only use monolith if we are set to
                matchrdirpath = next((match for match in matches if match.value == \
                                      theobj._resourcedir), None) if theobj.directory_load else None
                if not matchrdirpath and crawl:
                    for match in matches:
                        if path == "/rest/v1" and not loadcomplete:
                            if str(match.full_path) == "links.Schemas.href" or \
                                    str(match.full_path) == "links.Registries.href":
                                continue
                        elif not loadcomplete:
                            if str(match.full_path) == "Registries.@odata.id" or \
                                    str(match.full_path) == "JsonSchemas.@odata.id":
                                continue

                        if match.value == path:
                            continue
                        elif not isinstance(match.value, six.string_types):
                            continue

                        href = '%s' % match.value
                        theobj.get_queue.put((href, includelogs, loadcomplete, crawl, \
                                   rel, init, fpath(str(match.full_path), path), \
                                   originaluri, theobj))
                elif crawl:
                    href = '%s' % matchrdirpath.value
                    theobj.get_queue.put((href, includelogs, loadcomplete, crawl, \
                                   rel, init, path, originaluri, theobj))
                if loadcomplete:
                    if path == '/rest/v1':
                        schemamatch = jsonpath_rw.parse('$..extref')
                    else:
                        schemamatch = jsonpath_rw.parse('$..Uri')
                    smatches = schemamatch.find(resp.dict)
                    matches = matches + smatches
                    for match in matches:
                        if isinstance(match.value, six.string_types):
                            theobj.get_queue.put((match.value, includelogs, loadcomplete, 
                                crawl, rel, init, fpath(str(match.full_path), path), \
                                   originaluri, theobj))
                self.queue.task_done()
        except Empty:
            pass
        except Exception as excp:
            self.exception = excp

    def get_exception(self):
        """Get any exception from the thread"""
        return self.exception



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""Base implementation for interaction with blob store interface"""

#---------Imports---------

import os
import sys
import struct
import random
import string
import logging

from ctypes import c_char_p, c_ubyte, c_uint, cdll, POINTER, create_string_buffer, c_ushort

from redfish.hpilo.rishpilo import HpIlo, HpIloInitialError
from redfish.hpilo.rishpilo import BlobReturnCodes as hpiloreturncodes

if os.name == 'nt':
    from ctypes import windll
else:
    from _ctypes import dlclose

#---------End of imports---------
#---------Debug logger---------

LOGGER = logging.getLogger(__name__)

#---------End of debug logger---------
#-----------------------Error Returns----------------------

class UnexpectedResponseError(Exception):
    """Raise when we get data that we don't expect from iLO"""
    pass

class HpIloError(Exception):
    """Raised when iLO returns non-zero error code"""
    pass

class Blob2CreateError(Exception):
    """Raised when create operation fails"""
    pass

class Blob2InfoError(Exception):
    """Raised when create operation fails"""
    pass

class Blob2ReadError(Exception):
    """Raised when read operation fails"""
    pass

class Blob2WriteError(Exception):
    """Raised when write operation fails"""
    pass

class Blob2DeleteError(Exception):
    """Raised when delete operation fails"""
    pass

class Blob2OverrideError(Exception):
    """Raised when delete operation fails because of it been overwritten"""
    pass

class BlobRetriesExhaustedError(Exception):
    """Raised when max retries have been attempted for same operation"""
    pass

class Blob2FinalizeError(Exception):
    """Raised when finalize operation fails"""
    pass

class Blob2ListError(Exception):
    """Raised when list operation fails"""
    pass

class Blob2SecurityError(Exception):
    """Raised when there is an issue with security"""
    pass

class BlobNotFoundError(Exception):
    """Raised when blob not found in key/namespace"""
    pass

class ChifDllMissingError(Exception):
    """Raised when unable to obtain ilorest_chif dll handle"""
    pass

class EncryptionEnabledError(Exception):
    """Raised when high security encryption is enabled"""
    pass

#----------------------------------------------------------

#-------------------Helper functions-------------------------

class BlobReturnCodes(object):
    """Blob store return codes.

    SUCCESS           success
    BADPARAMETER      bad parameter supplied
    NOTFOUND          blob name not found
    NOTMODIFIED       call did not perform the operation

    """

    SUCCESS = 0
    BADPARAMETER = 2
    NOTFOUND = 12
    NOTMODIFIED = 20

class BlobStore2(object):
    """Blob store 2 class"""
    def __init__(self):
        lib = self.gethprestchifhandle()
        self.channel = HpIlo(dll=lib)
        self.max_retries = 3

    def __del__(self):
        """Blob store 2 close channel function"""
        if hasattr(self, 'channel'):
            self.channel.close()

    def create(self, key, namespace):
        """Create the blob

        :param key: The blob key to create.
        :type key: str.
        :param namespace: The blob namespace to create the key in.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.create_not_blobentry.argtypes = [c_char_p, c_char_p]
        lib.create_not_blobentry.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode('utf-8'))
        namespace = create_string_buffer(namespace.encode('utf-8'))

        ptr = lib.create_not_blobentry(name, namespace)
        data = ptr[:lib.size_of_createRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def get_info(self, key, namespace, retries=0):
        """Get information for a particular blob

        :param key: The blob key to retrieve.
        :type key: str.
        :param namespace: The blob namespace to retrieve the key from.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.get_info.argtypes = [c_char_p, c_char_p]
        lib.get_info.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode('utf-8'))
        namspace = create_string_buffer(namespace.encode('utf-8'))

        ptr = lib.get_info(name, namspace)
        data = ptr[:lib.size_of_infoRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if errorcode == BlobReturnCodes.BADPARAMETER:
            if retries < self.max_retries:
                self.get_info(key=key, namespace=namespace, retries=retries+1)
            else:
                raise Blob2OverrideError(errorcode)
        elif errorcode == BlobReturnCodes.NOTFOUND:
            raise BlobNotFoundError(key, namespace)

        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        response = resp[lib.size_of_responseHeaderBlob():]

        self.unloadchifhandle(lib)

        return response

    def read(self, key, namespace, retries=0):
        """Read a particular blob

        :param key: The blob key to be read.
        :type key: str.
        :param namespace: The blob namespace to read the key from.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        maxread = lib.max_read_size()
        readsize = lib.size_of_readRequest()
        readhead = lib.size_of_responseHeaderBlob()

        self.unloadchifhandle(lib)

        blob_info = self.get_info(key, namespace)
        blobsize = struct.unpack("<I", bytes(blob_info[0:4]))[0]

        bytes_read = 0
        data = bytearray()

        while bytes_read < blobsize:
            if (maxread - readsize) < (blobsize - bytes_read):
                count = maxread - readsize
            else:
                count = blobsize - bytes_read

            read_block_size = bytes_read
            recvpkt = self.read_fragment(key, namespace, read_block_size, count)

            newreadsize = readhead + 4
            bytesread = struct.unpack("<I", bytes(recvpkt[readhead:(newreadsize)]))[0]

            if bytesread == 0:
                if retries < self.max_retries:
                    data = self.read(key=key, namespace=namespace, retries=retries+1)
                    return data
                else:
                    raise BlobRetriesExhaustedError()

            data.extend(recvpkt[newreadsize:newreadsize + bytesread])
            bytes_read += bytesread

        return data

    def read_fragment(self, key, namespace, offset=0, count=1):
        """Fragmented version of read function for large blobs

        :param key: The blob key to be read.
        :type key: str.
        :param namespace: The blob namespace to read the key from.
        :type namespace: str.
        :param offset: The data offset for the current fragmented read.
        :type key: int.
        :param count: The data count for the current fragmented read.
        :type namespace: int.

        """
        lib = self.gethprestchifhandle()
        lib.read_fragment.argtypes = [c_uint, c_uint, c_char_p, c_char_p]
        lib.read_fragment.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode('utf-8'))
        namespace = create_string_buffer(namespace.encode('utf-8'))

        ptr = lib.read_fragment(offset, count, name, namespace)
        data = ptr[:lib.size_of_readRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        resp = resp + b"\0" * (lib.size_of_readResponse() - len(resp))

        return resp

    def write(self, key, namespace, data=None):
        """Write a particular blob

        :param key: The blob key to be written.
        :type key: str.
        :param namespace: The blob namespace to write the key in.
        :type namespace: str.
        :param data: The blob data to be written.
        :type data: str.

        """
        lib = self.gethprestchifhandle()
        maxwrite = lib.max_write_size()
        writesize = lib.size_of_writeRequest()

        self.unloadchifhandle(lib)

        if data:
            data_length = len(data)
            bytes_written = 0

            while bytes_written < data_length:
                if (maxwrite - writesize) < (data_length - bytes_written):
                    count = maxwrite - writesize
                else:
                    count = data_length - bytes_written

                write_blob_size = bytes_written

                self.write_fragment(key, namespace=namespace, \
                            data=data[write_blob_size:write_blob_size+count], \
                            offset=write_blob_size, count=count)

                bytes_written += count

        return self.finalize(key, namespace=namespace)

    def write_fragment(self, key, namespace, data=None, offset=0, count=1):
        """Fragmented version of write function for large blobs

        :param key: The blob key to be written.
        :type key: str.
        :param namespace: The blob namespace to write the key in.
        :type namespace: str.
        :param data: The blob data to be written to blob.
        :type data: str.
        :param offset: The data offset for the current fragmented write.
        :type key: int.
        :param count: The data count for the current fragmented write.
        :type count: int.

        """
        lib = self.gethprestchifhandle()
        lib.write_fragment.argtypes = [c_uint, c_uint, c_char_p, c_char_p]
        lib.write_fragment.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode('utf-8'))
        namespace = create_string_buffer(namespace.encode('utf-8'))

        ptr = lib.write_fragment(offset, count, name, namespace)
        sendpacket = ptr[:lib.size_of_writeRequest()]

        dataarr = bytearray(sendpacket)
        dataarr.extend(memoryview(data))

        resp = self._send_receive_raw(dataarr)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def delete(self, key, namespace, retries=0):
        """Delete the blob

        :param key: The blob key to be deleted.
        :type key: str.
        :param namespace: The blob namespace to delete the key from.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.delete_blob.argtypes = [c_char_p, c_char_p]
        lib.delete_blob.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode('utf-8'))
        namspace = create_string_buffer(namespace.encode('utf-8'))

        ptr = lib.delete_blob(name, namspace)
        data = ptr[:lib.size_of_deleteRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if errorcode == BlobReturnCodes.BADPARAMETER:
            if retries < self.max_retries:
                self.delete(key=key, namespace=namespace, retries=\
                                                    retries+1)
            else:
                raise Blob2OverrideError(errorcode)
        elif not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return errorcode

    def list(self, namespace):
        """List operation to retrieve all blobs in a given namespace

        :param namespace: The blob namespace to retrieve the keys from.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.list_blob.argtypes = [c_char_p]
        lib.list_blob.restype = POINTER(c_ubyte)

        namespace = create_string_buffer(namespace.encode('utf-8'))

        ptr = lib.list_blob(namespace)
        data = ptr[:lib.size_of_listRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        resp = resp + "\0" * (lib.size_of_listResponse() - len(resp))

        self.unloadchifhandle(lib)

        return resp

    def finalize(self, key, namespace):
        """Finalize the blob

        :param key: The blob key to be finalized.
        :type key: str.
        :param namespace: The blob namespace to finalize the key in.
        :type namespace: str.

        """
        lib = self.gethprestchifhandle()
        lib.finalize_blob.argtypes = [c_char_p, c_char_p]
        lib.finalize_blob.restype = POINTER(c_ubyte)

        name = create_string_buffer(key.encode('utf-8'))
        namespace = create_string_buffer(namespace.encode('utf-8'))

        ptr = lib.finalize_blob(name, namespace)
        data = ptr[:lib.size_of_finalizeRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return errorcode

    def rest_immediate(self, req_data, rqt_key="RisRequest", \
            rsp_key="RisResponse", rsp_namespace="volatile"):
        """Read/write blob via immediate operation

        :param req_data: The blob data to be read/written.
        :type req_data: str.
        :param rqt_key: The blob key to be used for the request data.
        :type rqt_key: str.
        :param rsp_key: The blob key to be used for the response data.
        :type rsp_key: str.
        :param rsp_namespace: The blob namespace to retrieve the response from.
        :type rsp_namespace: str.

        """
        rqt_key = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(10))
        rsp_key = ''.join(random.choice(string.ascii_letters + string.digits) for _ in range(10))

        lib = self.gethprestchifhandle()

        if len(req_data) < (lib.size_of_restImmediateRequest() + lib.max_write_size()):
            lib.rest_immediate.argtypes = [c_uint, c_char_p, c_char_p]
            lib.rest_immediate.restype = POINTER(c_ubyte)

            name = create_string_buffer(rsp_key.encode('utf-8'))
            namespace = create_string_buffer(rsp_namespace.encode('utf-8'))

            ptr = lib.rest_immediate(len(req_data), name, namespace)
            sendpacket = ptr[:lib.size_of_restImmediateRequest()]
            mode = False
        else:
            self.create(rqt_key, rsp_namespace)
            self.write(rqt_key, rsp_namespace, req_data)

            lib.rest_immediate_blobdesc.argtypes = [c_char_p, c_char_p, c_char_p]
            lib.rest_immediate_blobdesc.restype = POINTER(c_ubyte)

            name = create_string_buffer(rqt_key.encode('utf-8'))
            namespace = create_string_buffer(rsp_namespace.encode('utf-8'))
            rspname = create_string_buffer(rsp_key.encode('utf-8'))

            ptr = lib.rest_immediate_blobdesc(name, rspname, namespace)
            sendpacket = ptr[:lib.size_of_restBlobRequest()]
            mode = True

        data = bytearray(sendpacket)

        if not mode:
            data.extend(req_data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if errorcode == BlobReturnCodes.NOTFOUND:
            raise BlobNotFoundError(rsp_key, rsp_namespace)

        recvmode = struct.unpack("<I", bytes(resp[12:16]))[0]

        fixdlen = lib.size_of_restResponseFixed()
        response = resp[fixdlen:struct.unpack("<I", bytes(resp[16:20]))[0] + fixdlen]

        tmpresponse = None
        if errorcode == BlobReturnCodes.SUCCESS and not mode:
            if recvmode == 0:
                tmpresponse = ''.join(map(chr, response))
        elif errorcode == BlobReturnCodes.NOTMODIFIED and not mode:
            if recvmode == 0:
                tmpresponse = ''.join(map(chr, response))
        elif errorcode == BlobReturnCodes.SUCCESS:
            if recvmode == 0:
                tmpresponse = ''.join(map(chr, response))
        elif recvmode == 0:
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        if not tmpresponse and recvmode == 1:
            tmpresponse = self.read(rsp_key, rsp_namespace)

            try:
                self.delete(rsp_key, rsp_namespace)
            except Exception as excp:
                raise excp
        else:
            try:
                self.delete(rsp_key, rsp_namespace)
            except Blob2OverrideError as excp:
                pass
            except Exception as excp:
                raise excp

        return tmpresponse

    def get_security_state(self):
        """Get information for the current security state"""
        lib = self.gethprestchifhandle()
        lib.get_security_state.argtypes = []
        lib.get_security_state.restype = POINTER(c_ubyte)

        ptr = lib.get_security_state()
        data = ptr[:lib.size_of_securityStateRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = struct.unpack("<I", bytes(resp[8:12]))[0]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        try:
            securitystate = struct.unpack("<c", bytes(resp[72]))[0]
        except:
            securitystate = int(resp[72])

        self.unloadchifhandle(lib)

        return securitystate

    def mount_blackbox(self):
        """Operation to mount the blackbox partition"""
        lib = self.gethprestchifhandle()
        lib.blackbox_media_mount.argtypes = []
        lib.blackbox_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.blackbox_media_mount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def absaroka_media_mount(self):
        """Operation to mount the absaroka repo partition"""
        lib = self.gethprestchifhandle()
        lib.absaroka_media_mount.argtypes = []
        lib.absaroka_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.absaroka_media_mount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def gaius_media_mount(self):
        """Operation to mount the gaius media partition"""
        lib = self.gethprestchifhandle()
        lib.gaius_media_mount.argtypes = []
        lib.gaius_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.gaius_media_mount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def vid_media_mount(self):
        """Operation to mount the gaius media partition"""
        lib = self.gethprestchifhandle()
        lib.vid_media_mount.argtypes = []
        lib.vid_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.vid_media_mount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def mountflat(self):
        """Operation to mount the gaius media partition"""
        lib = self.gethprestchifhandle()
        lib.flat_media_mount.argtypes = []
        lib.flat_media_mount.restype = POINTER(c_ubyte)

        ptr = lib.flat_media_mount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.media_unmount.argtypes = []
        lib.media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.media_unmount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def bb_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.bb_media_unmount.argtypes = []
        lib.bb_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.bb_media_unmount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def vid_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.vid_media_unmount.argtypes = []
        lib.vid_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.vid_media_unmount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def gaius_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.gaius_media_unmount.argtypes = []
        lib.gaius_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.gaius_media_unmount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def absr_media_unmount(self):
        """Operation to unmount the media partition"""
        lib = self.gethprestchifhandle()
        lib.absaroka_media_unmount.argtypes = []
        lib.absaroka_media_unmount.restype = POINTER(c_ubyte)

        ptr = lib.absaroka_media_unmount()
        data = ptr[:lib.size_of_embeddedMediaRequest()]
        data = bytearray(data)

        resp = self._send_receive_raw(data)

        errorcode = resp[12]
        if not (errorcode == BlobReturnCodes.SUCCESS or errorcode == BlobReturnCodes.NOTMODIFIED):
            raise HpIloError(errorcode)

        self.unloadchifhandle(lib)

        return resp

    def _send_receive_raw(self, indata):
        """Send and receive raw function for blob operations

        :param indata: The data to be sent to blob operation.
        :type indata: str.

        """
        excp = None
        for _ in range(0, 3): # channel loop for iLO
            try:
                resp = self.channel.send_receive_raw(indata, 10)
                return resp
            except Exception as exp:
                self.channel.close()
                lib = self.gethprestchifhandle()
                self.channel = HpIlo(dll=lib)
                excp = exp
        if excp:
            raise excp

    @staticmethod
    def gethprestchifhandle():
        """Multi platform handle for chif hprest library"""
        excp = None
        libhandle = None
        libnames = ["ilorest_chif.dll", "hprest_chif.dll"] if os.name == \
                    'nt' else ["ilorest_chif_dev.so", "hprest_chif_dev.so", \
                                        "ilorest_chif.so", "hprest_chif.so"]
        for libname in libnames:
            try:
                libpath = BlobStore2.checkincurrdirectory(libname)
                libhandle = cdll.LoadLibrary(libpath)
                if libhandle:
                    break
            except Exception as exp:
                excp = exp

        if libhandle:
            BlobStore2.setglobalhprestchifrandnumber(libhandle)
            return libhandle
        raise ChifDllMissingError(excp)

    @staticmethod
    def setglobalhprestchifrandnumber(libbhndl):
        """Set the random number for the chif handle
        :param libbhndl: The library handle provided by loading the chif library.
        :type libbhndl: library handle.
        """
        rndval = random.randint(1, 65535)
        libbhndl.updaterandval.argtypes = [c_ushort]
        libbhndl.updaterandval(rndval)

    @staticmethod
    def initializecreds(username=None, password=None):
        """Get chif ready to use high security
        :param username: The username to login.
        :type username: str.
        :param password: The password to login.
        :type password: str.
        """
        #TODO: initialize credentials for certificate
        dll = BlobStore2.gethprestchifhandle()
        if LOGGER.isEnabledFor(logging.DEBUG):
            dll.enabledebugoutput()
        dll.ChifInitialize(None)
        if dll.ChifIsSecurityRequired() > 0:
            if not username or not password:
                return False
            dll.initiate_credentials.argtypes = [c_char_p, c_char_p]
            dll.initiate_credentials.restype = POINTER(c_ubyte)

            usernew = create_string_buffer(username.encode('utf-8'))
            passnew = create_string_buffer(password.encode('utf-8'))

            dll.initiate_credentials(usernew, passnew)
            credreturn = dll.ChifVerifyCredentials()
            if not credreturn == BlobReturnCodes.SUCCESS:
                if credreturn == hpiloreturncodes.CHIFERR_AccessDenied:
                    raise Blob2SecurityError()
                else:
                    raise HpIloInitialError("Error %s occurred while trying " \
                                        "to open a channel to iLO" % credreturn)
        else:
            #so we don't have extra overhead if we don't have to
            dll.ChifDisableSecurity()
        BlobStore2.unloadchifhandle(dll)

        return True

    @staticmethod
    def checkincurrdirectory(libname):
        """Check if the library is present in current directory.
        :param libname: The name of the library to search for.
        :type libname: str."""
        libpath = libname

        if os.path.isfile(os.path.join(os.path.split(sys.executable)[0], libpath)):
            libpath = os.path.join(os.path.split(sys.executable)[0], libpath)
        elif os.path.isfile(os.path.join(os.getcwd(), libpath)):
            libpath = os.path.join(os.getcwd(), libpath)
        elif os.environ.get("LD_LIBRARY_PATH"):
            paths = os.getenv("LD_LIBRARY_PATH", libpath).split(';')
            libpath = [os.path.join(pat, libname) for pat in paths if \
                                                        os.path.isfile(os.path.join(pat, libname))]
            libpath = libpath[0] if libpath else libname

        return libpath

    @staticmethod
    def unloadchifhandle(lib):
        """Release a handle on the chif iLOrest library

        :param lib: The library handle provided by loading the chif library.
        :type lib: library handle.

        """
        try:
            libhandle = lib._handle
            if os.name == 'nt':
                windll.kernel32.FreeLibrary(None, handle=libhandle)
            else:
                dlclose(libhandle)
        except Exception:
            pass



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""RMC helper file. Includes RMC errors and caching functionality for monolith."""

#---------Imports---------

import os
import json
import errno
import logging
import hashlib

from redfish.rest.v1 import RestClient
from redfish.rest.containers import StaticRestResponse, RestRequest

from .ris import (RisMonolith)
from .sharedtypes import (JSONEncoder)

#---------End of imports---------

#---------Debug logger---------

LOGGER = logging.getLogger(__name__)

#---------End of debug logger---------

class RdmcError(Exception):
    """Base class for all RDMC Exceptions"""
    errcode = 1
    def __init__(self, message):
        Exception.__init__(self, message)

class InvalidCommandLineError(RdmcError):
    """Raised when user enter incorrect command line arguments"""
    pass

class FailureDuringCommitError(RdmcError):
    """Raised when there is an error while committing."""
    pass

class UserNotAdminError(RdmcError):
    """Raised when user doesn't have admin priviledges, but they are required."""
    pass

class UndefinedClientError(Exception):
    """Raised when there are no clients active (usually when user hasn't logged in)."""
    pass

class InstanceNotFoundError(Exception):
    """Raised when attempting to select an instance that does not exist."""
    pass

class CurrentlyLoggedInError(Exception):
    """Raised when attempting to select an instance that does not exist"""
    pass

class NothingSelectedError(Exception):
    """Raised when attempting to access an object without first selecting it."""
    pass

class NothingSelectedFilterError(Exception):
    """Raised when the filter applied doesn't match any selection (general)."""
    pass

class NothingSelectedSetError(Exception):
    """Raised when attempting to access an object without first selecting it (In set)."""
    pass

class InvalidSelectionError(Exception):
    """Raised when selection argument fails to match anything."""
    pass

class IdTokenError(Exception):
    """Raised when user is not authorized to complete the operation."""
    pass

class ValueChangedError(Exception):
    """Raised if user tries to set/commit a value when monolith has older data."""
    pass

class LoadSkipSettingError(Exception):
    """Raised when one or more settings are absent in given server."""
    pass

class InvalidPathError(Exception):
    """Raised when requested path is not found."""
    pass

class UnableToObtainIloVersionError(Exception):
    """Raised when iloversion is missing from default path."""
    pass

class IncompatibleiLOVersionError(Exception):
    """Raised when the iLO version is above or below the required version."""
    pass

class ValidationError(Exception):
    """Raised when there is a problem with user input."""
    def __init__(self, errlist):
        super(ValidationError, self).__init__(errlist)
        self._errlist = errlist

    def get_errors(self):
        """Returns error list."""
        return self._errlist

class IloResponseError(Exception):
    """Raised when iLO returns with a non 2XX response."""
    pass

class EmptyRaiseForEAFP(Exception):
    """Raised when you need to check for issues and take different action."""
    pass

class IncorrectPropValue(Exception):
    """Raised when you pass an incorrect value to for the associated property."""
    pass

class RmcCacheManager(object):
    """Manages caching/uncaching of data for RmcApp.

    :param rmc: RmcApp to be managed
    :type rmc: :class:`redfish.ris.rmc.RmcApp`
    """
    def __init__(self, rmc):
        """Initialize RmcCacheManager

        :param rmc: RmcApp to be managed
        :type rmc: RmcApp object

        """
        self._rmc = rmc

        self.encodefunct = lambda data: data
        self.decodefunct = lambda data: data

class RmcFileCacheManager(RmcCacheManager):
    """RMC file cache manager.

    :param rmc: RmcApp to be managed
    :type rmc: :class:`redfish.ris.rmc.RmcApp`
    """
    def __init__(self, rmc):
        super(RmcFileCacheManager, self).__init__(rmc)

    def logout_del_function(self, url=None):
        """Searches for a specific url in cache or returns all urls and returns them for RmcApp
        to run logout on, clearing the session.

        :param url: The URL to pass back for logout.
        :type url: str
        """
        if self._rmc.cache:
            cachedir = self._rmc.cachedir
            indexfn = os.path.join(cachedir, "index") #%s\\index' % cachedir
        else:
            indexfn = ''
        sessionlocs = []

        if os.path.isfile(indexfn):
            try:
                indexfh = open(indexfn, 'r')
                index_cache = json.load(indexfh)
                indexfh.close()

                for index in index_cache:
                    if url:
                        if url in index['url']:
                            os.remove(os.path.join(cachedir, index['href']))
                            break
                    else:
                        if os.path.isfile(os.path.join(cachedir, index['href'])):
                            monolith = open(os.path.join(cachedir, index['href']), 'r')
                            data = json.load(monolith)
                            monolith.close()
                            for item in data:
                                if 'login' in item and 'session_location' in data['login']:
                                    if 'blobstore' in data['login']['url']:
                                        loc = data['login']['session_location']\
                                                .split('//')[-1]
                                        sesurl = None
                                    else:
                                        loc = data['login']['session_location']\
                                                .split(data['login']['url'])[-1]
                                        sesurl = data['login']['url']
                                    sessionlocs.append((loc, sesurl,\
                                                self._rmc._cm.decodefunct\
                                                (data['login']['session_key'])))

                        os.remove(os.path.join(cachedir, index['href']))
            except BaseException as excp:
                LOGGER.warning('Unable to read cache data %s', excp)

        return sessionlocs

    def uncache_rmc(self, creds=None, enc=False):
        """Uncaches monolith data from cache location specified by RmcApp.

        :param creds: Dictionary of username and password.
                      Only required for restoring high security local calls.
        :type creds: dict
        :param enc: Flag if credentials passed are encoded.
        :type enc: bool
        """
        cachedir = self._rmc.cachedir
        indexfn = '%s/index' % cachedir

        if os.path.isfile(indexfn):
            try:
                indexfh = open(indexfn, 'r')
                index_cache = json.load(indexfh)
                indexfh.close()

                for index in index_cache:
                    clientfn = index['href']
                    self._uncache_client(clientfn, creds=creds, enc=enc)
            except BaseException as excp:
                LOGGER.warning('Unable to read cache data %s', excp)

    def _uncache_client(self, cachefn, creds=None, enc=False):
        """Complex monolith uncache function

        :param cachefn: The cache file name.
        :type cachefn: str.

        """
        cachedir = self._rmc.cachedir
        clientsfn = '%s/%s' % (cachedir, cachefn)

        if os.path.isfile(clientsfn):
            try:
                clientsfh = open(clientsfn, 'r')
                client = json.load(clientsfh)
                clientsfh.close()

                if 'login' not in client:
                    return

                login_data = client['login']
                if 'url' not in login_data:
                    return

                self._rmc.typepath.getgen(login_data.get('ilo'), \
                                 url=login_data.get('url'), \
                                 isredfish=login_data.get('redfish', None), \
                                 ca_certs=login_data.get('ca_certs', None))
                #TODO: Add certificate data
                if creds and login_data.get('url', '').startswith('blobstore://'):
                    if enc:
                        creds['username'] = self._rmc._cm.decodefunct(creds['username'])
                        creds['password'] = self._rmc._cm.decodefunct(creds['password'])
                    login_data['username'] = creds['username']
                    login_data['password'] = creds['password']

                redfishinst = RestClient(\
                    username=login_data.get('username', 'Administrator'), \
                    password=login_data.get('password', None), \
                    base_url=login_data.get('url', None), \
                    biospassword=login_data.get('bios_password', None), \
                    is_redfish=login_data.get('redfish', None), \
                    default_prefix=self._rmc.typepath.defs.startpath, \
                    proxy=login_data.get('proxy', None), \
                    ca_certs=login_data.get('ca_certs', None))
                if login_data.get('authorization_key'):
                    redfishinst.basic_auth = login_data.get('authorization_key')
                elif login_data.get('session_key'):
                    redfishinst.session_key = \
                                            self._rmc._cm.decodefunct(login_data.get('session_key'))
                    redfishinst.session_location = login_data.get('session_location')

                if 'selector' in client:
                    self._rmc.selector = client['selector']

                getdata = client['get']
                for key in list(getdata.keys()):
                    if key == redfishinst.default_prefix:
                        restreq = RestRequest(method='GET', path=key)
                        getdata[key]['restreq'] = restreq
                        redfishinst.root = StaticRestResponse(**getdata[key])
                        break

                self._rmc.monolith = RisMonolith(redfishinst, self._rmc.typepath)
                self._rmc.monolith.load_from_dict(client['monolith'])
                self._rmc.redfishinst = redfishinst
                #make sure root is there
                _ = redfishinst.root
                self._rmc.typepath.defineregschemapath(redfishinst.root.dict)
            except BaseException as excp:
                LOGGER.warning('Unable to read cache data %s', excp)

    def cache_rmc(self):
        """Saves monolith data to the file path specified in RmcApp."""
        if not self._rmc.cache:
            return

        cachedir = self._rmc.cachedir
        if not os.path.isdir(cachedir):
            try:
                os.makedirs(cachedir)
            except OSError as ex:
                if ex.errno == errno.EEXIST:
                    pass
                else:
                    raise

        index_map = dict()
        index_cache = list()

        if self._rmc.redfishinst:
            shaobj = hashlib.new("SHA256")
            shaobj.update(self._rmc.redfishinst.base_url.encode('utf-8'))
            md5str = shaobj.hexdigest()

            index_map[self._rmc.redfishinst.base_url] = md5str
            index_data = dict(url=self._rmc.redfishinst.base_url, href='%s' % md5str,)
            index_cache.append(index_data)

            indexfh = open('%s/index' % cachedir, 'w')
            json.dump(index_cache, indexfh, indent=2, cls=JSONEncoder)
            indexfh.close()

        if self._rmc.redfishinst:
            #TODO: Add certificate data
            login_data = dict(\
                username=None, \
                password=None, url=self._rmc.redfishinst.base_url, \
                session_key=self._rmc._cm.encodefunct(self._rmc.redfishinst.session_key), \
                session_location=self._rmc.redfishinst.session_location, \
                authorization_key=self._rmc.redfishinst.basic_auth, \
                bios_password=self._rmc.redfishinst.bios_password, \
                redfish=self._rmc.monolith.is_redfish, \
                ilo=self._rmc.typepath.ilogen,\
                proxy=self._rmc.redfishinst.proxy,\
                ca_certs=self._rmc.redfishinst.connection.\
                                                        _connection_properties.get('ca_certs',None))

            clients_data = dict(selector=self._rmc.selector, login=login_data, \
                     monolith=self._rmc.monolith, get=self._rmc.monolith.paths)

            clientsfh = open('%s/%s' % (cachedir, \
                                     index_map[self._rmc.redfishinst.base_url]), 'w')

            json.dump(clients_data, clientsfh, indent=2, cls=JSONEncoder)
            clientsfh.close()



# -*- coding: utf-8 -*-
"""
Expanded LegacyREST/Redfish interface for schema validation, database for responses, caching,
and error registries.
"""

from .sharedtypes import (
    JSONEncoder
)

from .ris import (
    RisInstanceNotFoundError,
    RisMonolithMemberBase,
    RisMonolithMemberv100,
    RisMonolith,
    SessionExpired,
)

from .rmc_helper import (
    UndefinedClientError,
    InstanceNotFoundError,
    CurrentlyLoggedInError,
    NothingSelectedError,
    NothingSelectedFilterError,
    NothingSelectedSetError,
    InvalidSelectionError,
    IdTokenError,
    ValidationError,
    ValueChangedError,
    RmcCacheManager,
    RmcFileCacheManager,
)

from .rmc import (
    RmcApp
)

from .validation import (
    ValidationManager,
    RegistryValidationError
)



 # Copyright 2020 Hewlett Packard Enterprise Development LP
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

# -*- coding: utf-8 -*-
"""
An example of gathering the computer system details
"""


import sys
import json
from redfish import RedfishClient
from redfish.rest.v1 import ServerDownOrUnreachableError

from get_resource_directory import get_resource_directory

def computer_details(_redfishobj):
    systems_members_uri = None
    systems_members_response = None

    resource_instances = get_resource_directory(_redfishobj)
    if DISABLE_RESOURCE_DIR or not resource_instances:
        #if we do not have a resource directory or want to force it's non use to find the
        #relevant URI
        systems_uri = _redfishobj.root.obj['Systems']['@odata.id']
        systems_response = _redfishobj.get(systems_uri)
        systems_members_uri = next(iter(systems_response.obj['Members']))['@odata.id']
        systems_members_response = _redfishobj.get(systems_members_uri)
    else:
        for instance in resource_instances:
            #Use Resource directory to find the relevant URI
            if '#ComputerSystem.' in instance['@odata.type']:
                systems_members_uri = instance['@odata.id']
                systems_members_response = _redfishobj.get(systems_members_uri)

    print("\n\nPrinting computer system details:\n\n")
    print(json.dumps(systems_members_response.dict, indent=4, sort_keys=True))

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the secured (https://) address,
    # account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://16.83.61.9"
    LOGIN_ACCOUNT = "Administrator"
    LOGIN_PASSWORD = "password"

    # flag to force disable resource directory. Resource directory and associated operations are
    # intended for HPE servers.
    DISABLE_RESOURCE_DIR = False

    try:
        # Create a Redfish client object
        REDFISHOBJ = RedfishClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, \
                                                                            password=LOGIN_PASSWORD)
        # Login with the Redfish client
        REDFISHOBJ.login()
    except ServerDownOrUnreachableError as excp:
        sys.stderr.write("ERROR: server not reachable or does not support RedFish.\n")
        sys.exit()

    computer_details(REDFISHOBJ)
    REDFISHOBJ.logout()



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.

import sys
from redfish import LegacyRestClient
from get_resource_directory import get_resource_directory

def update_ilo_firmware(restobj, fw_url=None, tpm_flag=False):
    resource_instances = get_resource_directory(restobj)
    if resource_instances:
        #Get URI from resource directory
        for instance in resource_instances:
            if "HpiLOFirmwareUpdate." in instance.Type:
                update_path = instance.href
                break
    body = dict()
    body["Action"] = "InstallFromURI"
    body["FirmwareURI"] = fw_url
    body["TPMOverrideFlag"] = tpm_flag
    response = restobj.post(update_path, body)
    sys.stdout.write("%s" % response)

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # While this example can be run remotely, it is used locally to locate the
    # iLO IP address
    #SYSTEM_URL = None
    #LOGIN_ACCOUNT = None
    #LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    REST_OBJ.login()
    sys.stdout.write("\nEXAMPLE 53: Update iLO Firmware\n")
    update_ilo_firmware(REST_OBJ, "http://test.com/ilo4_244.bin", False)
    REST_OBJ.logout()



from setuptools import setup, find_packages

extras = {'socks': ['pysocks']}

setup(name='python-ilorest-library',
      version='3.1.1',
      description='iLO Restful Python Library',
	  author = 'Hewlett Packard Enterprise',
	  author_email = 'rajeevalochana.kallur@hpe.com',
      extras_require = extras,
      classifiers=[
          'Development Status :: 5 - Production/Stable',
          'License :: OSI Approved :: Apache Software License',
          'Programming Language :: Python :: 2.7',
          'Programming Language :: Python :: 3',
          'Topic :: Communications'
      ],
      keywords='Hewlett Packard Enterprise iLORest',
      url='https://github.com/HewlettPackard/python-ilorest-library',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      install_requires=[
          'jsonpatch',
          'jsonpath_rw',
          'jsonpointer',
          'urllib3',
          'six'
      ])



 # Copyright 2020 Hewlett Packard Enterprise Development, LP.
 #
 # Licensed under the Apache License, Version 2.0 (the "License"); you may
 # not use this file except in compliance with the License. You may obtain
 # a copy of the License at
 #
 #      http://www.apache.org/licenses/LICENSE-2.0
 #
 # Unless required by applicable law or agreed to in writing, software
 # distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 # WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 # License for the specific language governing permissions and limitations
 # under the License.


"""
An example of adding a user account by iLO privileges
"""


import sys
from redfish import LegacyRestClient
from get_resource_directory import get_resource_directory

def add_ilo_user_account(restobj, new_ilo_loginname, new_ilo_username, \
                                 new_ilo_password, irc=False, cfg=False, \
                                 virtual_media=False, usercfg=False, vpr=False):
    resource_instances = get_resource_directory(restobj)
    if resource_instances:
        #Get URI from resource directory
        for instance in resource_instances:
            if "Collection." in instance.Type:
                if instance.MemberType.startswith("ManagerAccount."):
                    accounts_path = instance.href
                    break

    body = {"UserName": new_ilo_loginname, "Password": new_ilo_password, "Oem": {}}
    body["Oem"]["Hp"] = {}
    body["Oem"]["Hp"]["LoginName"] = new_ilo_username
    body["Oem"]["Hp"]["Privileges"] = {}
    body["Oem"]["Hp"]["Privileges"]["RemoteConsolePriv"] = irc
    body["Oem"]["Hp"]["Privileges"]["iLOConfigPriv"] = cfg
    body["Oem"]["Hp"]["Privileges"]["VirtualMediaPriv"] = virtual_media
    body["Oem"]["Hp"]["Privileges"]["UserConfigPriv"] = usercfg
    body["Oem"]["Hp"]["Privileges"]["VirtualPowerAndResetPriv"] = vpr

    response = restobj.post(accounts_path, body)

    sys.stdout.write("%s" % response)

if __name__ == "__main__":
    # When running on the server locally use the following commented values
    # SYSTEM_URL = None
    # LOGIN_ACCOUNT = None
    # LOGIN_PASSWORD = None

    # When running remotely connect using the iLO secured (https://) address,
    # iLO account name, and password to send https requests
    # SYSTEM_URL acceptable examples:
    # "https://10.0.0.100"
    # "https://ilo.hostname"
    SYSTEM_URL = "https://10.0.0.100"
    LOGIN_ACCOUNT = "admin"
    LOGIN_PASSWORD = "password"

    #Create a REST object
    REST_OBJ = LegacyRestClient(base_url=SYSTEM_URL, username=LOGIN_ACCOUNT, password=LOGIN_PASSWORD)
    REST_OBJ.login()
    sys.stdout.write("\nEXAMPLE 10: Create an iLO User Account\n")
    add_ilo_user_account(REST_OBJ, "name", "username", "password")
    REST_OBJ.logout()



###
# Copyright 2020 Hewlett Packard Enterprise, Inc. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
###

# -*- coding: utf-8 -*-
"""All Connections for interacting with REST."""
import time
import gzip
import json
import logging
import urllib3

from urllib3 import ProxyManager, PoolManager
from urllib3.exceptions import MaxRetryError, DecodeError

try:
    urllib3.disable_warnings()
    from urllib3.contrib.socks import SOCKSProxyManager
except ImportError:
    pass

import six
from six import BytesIO
from six.moves.urllib.parse import urlparse, urlencode

from redfish.hpilo.rishpilo import HpIloChifPacketExchangeError
from redfish.hpilo.risblobstore2 import BlobStore2, Blob2OverrideError, Blob2SecurityError
from redfish.rest.containers import RestRequest, RestResponse, RisRestResponse

#---------End of imports---------


#---------Debug logger---------

LOGGER = logging.getLogger(__name__)

#---------End of debug logger---------

class RetriesExhaustedError(Exception):
    """Raised when retry attempts have been exhausted."""
    pass

class DecompressResponseError(Exception):
    """Raised when decompressing the response failed."""
    pass

class InvalidCredentialsError(Exception):
    """Raised when invalid credentials have been provided."""
    pass

class ChifDriverMissingOrNotFound(Exception):
    """Raised when CHIF driver is missing or not found."""
    pass

class SecurityStateError(Exception):
    """Raised when there is a strict security state without authentication."""
    pass

class HttpConnection(object):
    """HTTP connection capable of authenticating with HTTPS and Http/Socks Proxies

    :param base_url: The URL to make HTTP calls against
    :type base_url: str
    :param \\**client_kwargs: Arguments to pass to the connection initialization. These are """\
    "passed to a urllib3 `PoolManager <https://urllib3.readthedocs.io/en/latest/reference/"\
    "index.html?highlight=PoolManager#urllib3.PoolManager>`_. All arguments that can be passed to "\
    "a PoolManager are valid arguments."
    def __init__(self, base_url, **client_kwargs):
        self._conn = None
        self.base_url = base_url
        self._connection_properties = client_kwargs
        self._proxy = self._connection_properties.pop('proxy', None)
        self.session_key = self._connection_properties.pop('session_key', None)
        self._init_connection()

    @property
    def proxy(self):
        """The proxy, if any."""
        return self._proxy

    @proxy.setter
    def proxy(self, proxy):
        """set the proxy"""
        self._proxy = proxy

    def _init_connection(self):
        """Function for initiating connection with remote server"""
        if self._connection_properties.get('ca_certs', None):
            LOGGER.info('Using CA cert to confirm identity.')
            cert_reqs = 'CERT_REQUIRED'
        else:
            LOGGER.info('Not using CA certificate.')
            cert_reqs = 'CERT_NONE'
        if self.proxy:
            if self.proxy.startswith('socks'):
                LOGGER.info("Initializing a SOCKS proxy.")
                http = SOCKSProxyManager(self.proxy, cert_reqs=cert_reqs, maxsize=6, \
                                                                    **self._connection_properties)
            else:
                LOGGER.info("Initializing a HTTP proxy.")
                http = ProxyManager(self.proxy, cert_reqs=cert_reqs, maxsize=6, \
                                    **self._connection_properties)
        else:
            LOGGER.info("Initializing no proxy.")
            http = PoolManager(cert_reqs=cert_reqs, maxsize=6, **self._connection_properties)

        self._conn = http.request

    def rest_request(self, path, method='GET', args=None, body=None, headers=None):
        """Format and do HTTP Rest request

        :param path: The URI path to perform the operation on.
        :type path: str
        :param method: method to perform on the path.
        :type method: str
        :param args: Any query to add to the URI. (Can also be directly added to the URI)
        :type args: dict
        :param body: body payload to include in the request if needed.
        :type body: dict
        :param headers: Any extra headers to add to the request.
        :type headers: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        #TODO: Need to remove redfish.dmtf.org calls from here, add to their own HttpConnection
        files = None
        request_args = {}
        external_uri = True if 'redfish.dmtf.org' in path else False
        headers = {} if external_uri else headers
        reqpath = path.replace('//', '/') if not external_uri else path

        if body is not None:
            if body and isinstance(body, list) and isinstance(body[0], tuple):
                files = body
                body = None
            elif isinstance(body, (dict, list)):
                headers['Content-Type'] = 'application/json'
                body = json.dumps(body)
            elif not files:
                headers['Content-Type'] = 'application/octet-stream'

            if method == 'PUT':
                resp = self.rest_request(method='HEAD', path=path, headers=headers)

                try:
                    if resp.getheader('content-encoding') == 'gzip':
                        buf = BytesIO()
                        gfile = gzip.GzipFile(mode='wb', fileobj=buf)

                        try:
                            gfile.write(str(body).encode('utf-8') if six.PY3 \
                                        else str(body))
                        finally:
                            gfile.close()

                        compresseddata = buf.getvalue()
                        if compresseddata:
                            data = bytearray()
                            data.extend(memoryview(compresseddata))
                            body = data
                except BaseException as excp:
                    LOGGER.error('Error occur while compressing body: %s', excp)
                    raise

        if args:
            if method == 'GET':
                reqpath += '?' + urlencode(args)
            elif method == 'PUT' or method == 'POST' or method == 'PATCH':
                headers['Content-Type'] = 'application/x-www-form-urlencoded'
                body = urlencode(args)

        #TODO: ADD to the default headers?
        headers['Accept-Encoding'] = 'gzip'

        restreq = RestRequest(path, method, data=files if files else body, url=self.base_url)

        if LOGGER.isEnabledFor(logging.DEBUG):
            try:
                logbody = None
                if restreq.body:
                    if restreq.body[0] == '{':
                        logbody = restreq.body
                    else:
                        raise KeyError()
                if restreq.method in ['POST', 'PATCH']:
                    debugjson = json.loads(restreq.body)
                    if 'Password' in debugjson.keys():
                        debugjson['Password'] = '******'
                    if 'OldPassword' in debugjson.keys():
                        debugjson['OldPassword'] = '******'
                    if 'NewPassword' in debugjson.keys():
                        debugjson['NewPassword'] = '******'
                    logbody = json.dumps(debugjson)
                LOGGER.debug('HTTP REQUEST: %s\n\tPATH: %s\n\t'\
                            'HEADERS: %s\n\tBODY: %s', restreq.method, restreq.path, headers, \
                             logbody)
            except:
                LOGGER.debug('HTTP REQUEST: %s\n\tPATH: %s\n\tBODY: %s', restreq.method, \
                                                            restreq.path, 'binary body')

        inittime = time.time()
        reqfullpath = self.base_url+reqpath if not external_uri else reqpath

        #To ensure we don't have unicode/string merging issues in httplib of Python 2
        if isinstance(reqfullpath, six.text_type):
            reqfullpath = str(reqfullpath)

        request_args['headers'] = headers
        if files:
            request_args['fields'] = files
        else:
            request_args['body'] = body
        try:
            resp = self._conn(method, reqfullpath, **request_args)
        except MaxRetryError:
            #TODO: Update error to be more descriptive
            raise RetriesExhaustedError()
        except DecodeError:
            #TODO: Update to be more descriptive or replace with DecodeError
            raise DecompressResponseError()

        endtime = time.time()
        LOGGER.info('Response Time to %s: %s seconds.', restreq.path, str(endtime-inittime))

        restresp = RestResponse(restreq, resp)

        if LOGGER.isEnabledFor(logging.DEBUG):
            headerstr = ''
            if restresp is not None:
                respheader = restresp.getheaders()
                for kiy, headerval in respheader.items():
                    headerstr += '\t' + kiy + ': ' + headerval + '\n'
                try:
                    LOGGER.debug('HTTP RESPONSE for %s:\nCode:%s\nHeaders:'\
                            '\n%s\nBody Response of %s: %s', restresp.request.path,\
                            str(restresp._http_response.status)+ ' ' + \
                            restresp._http_response.reason, \
                            headerstr, restresp.request.path, restresp.read\
                            .encode('ascii', 'ignore'))
                except:
                    LOGGER.debug('HTTP RESPONSE:\nCode:%s', restresp)
            else:
                LOGGER.debug('HTTP RESPONSE: No HTTP Response obtained')

        return restresp

    def cert_login(self):
        """Login using a certificate **NOT IMPLEMENTED**"""
        #TODO: Implement
        return self

class Blobstore2Connection(object):
    """A connection for communicating locally with HPE servers

    :param \\**conn_kwargs: Arguments to pass to the connection initialization.

    Possible arguments for *\\**conn_kwargs* include:

    :username: The username to login with
    :password: The password to login with
    """
    _http_vsn_str = 'HTTP/1.1'
    blobstore_headers = {'Accept': '*/*', 'Connection': 'Keep-Alive'}

    def __del__(self):
        """Clear channel"""
        self._conn = None

    def __init__(self, **conn_kwargs):
        self._conn = None
        self.base_url = "blobstore://."
        self._connection_properties = conn_kwargs
        self.session_key = self._connection_properties.pop('session_key', None)
        self._init_connection(**self._connection_properties)

    def _init_connection(self, **kwargs):
        """Initiate blobstore connection"""
        # mixed security modes require a password at all times
        username = kwargs.pop('username', 'nousername')
        password = kwargs.pop('password', 'nopassword')
        try:
            correctcreds = BlobStore2.initializecreds(username=username, password=password)
            bs2 = BlobStore2()
            if not correctcreds:
                security_state = int(bs2.get_security_state())
                raise SecurityStateError(security_state)
        except Blob2SecurityError:
            raise InvalidCredentialsError(0)
        except HpIloChifPacketExchangeError as excp:
            LOGGER.info("Exception: %s", str(excp))
            raise ChifDriverMissingOrNotFound()
        except Exception as excp:
            if str(excp) == 'chif':
                raise ChifDriverMissingOrNotFound()
            else:
                raise
        else:
            self._conn = bs2

    def rest_request(self, path='', method="GET", args=None, body=None, headers=None):
        """Rest request for blobstore client

        :param path: The URI path to perform the operation on.
        :type path: str
        :param method: method to perform on the path.
        :type method: str
        :param args: Any query to add to the URI. (Can also be directly added to the URI)
        :type args: dict
        :param body: body payload to include in the request if needed.
        :type body: dict
        :param headers: Any extra headers to add to the request.
        :type headers: dict
        :returns: A :class:`redfish.rest.containers.RestResponse` object
        """
        # default headers if not passed in - otherwise will throw on .update call
        if headers is None:
            headers = {}
        else:
            headers.update(Blobstore2Connection.blobstore_headers)

        reqpath = path.replace('//', '/')

        oribody = body
        if body is not None:
            if isinstance(body, (dict, list)):
                headers['Content-Type'] = 'application/json'
                body = json.dumps(body)
            else:
                headers['Content-Type'] = 'application/x-www-form-urlencoded'
                body = urlencode(body)

            if method == 'PUT':
                resp = self.rest_request(path=path, headers=headers)

                try:
                    if resp.getheader('content-encoding') == 'gzip':
                        buf = BytesIO()
                        gfile = gzip.GzipFile(mode='wb', fileobj=buf)

                        try:
                            gfile.write(str(body).encode('utf-8') if six.PY3 \
                                        else str(body))
                        finally:
                            gfile.close()

                        compresseddata = buf.getvalue()
                        if compresseddata:
                            data = bytearray()
                            data.extend(memoryview(compresseddata))
                            body = data
                except BaseException as excp:
                    LOGGER.error('Error occur while compressing body: %s', excp)
                    raise

            headers['Content-Length'] = len(body)

        if args:
            if method == 'GET':
                reqpath += '?' + urlencode(args)
            elif method == 'PUT' or method == 'POST' or method == 'PATCH':
                headers['Content-Type'] = 'application/x-www-form-urlencoded'
                body = urlencode(args)

        str1 = '{} {} {}\r\n'.format(method, reqpath, Blobstore2Connection._http_vsn_str)
        str1 += 'Host: \r\n'
        str1 += 'Accept-Encoding: gzip\r\n'
        for header, value in headers.items():
            str1 += '{}: {}\r\n'.format(header, value)

        str1 += '\r\n'

        if body and len(body) > 0:
            if isinstance(body, bytearray):
                str1 = bytearray(str1.encode("ASCII")) + body
            else:
                str1 += body

        if not isinstance(str1, bytearray):
            str1 = bytearray(str1.encode("ASCII"))

        if LOGGER.isEnabledFor(logging.DEBUG):
            try:
                logbody = None
                if body:
                    if body[0] == '{':
                        logbody = body
                    else:
                        raise
                if method in ['POST', 'PATCH']:
                    debugjson = json.loads(body)
                    if 'Password' in debugjson.keys():
                        debugjson['Password'] = '******'
                    if 'OldPassword' in debugjson.keys():
                        debugjson['OldPassword'] = '******'
                    if 'NewPassword' in debugjson.keys():
                        debugjson['NewPassword'] = '******'
                    logbody = json.dumps(debugjson)

                LOGGER.debug('Blobstore REQUEST: %s\n\tPATH: %s\n\tHEADERS: '\
                             '%s\n\tBODY: %s', method, str(headers), path, logbody)
            except:
                LOGGER.debug('Blobstore REQUEST: %s\n\tPATH: %s\n\tHEADERS: '\
                             '%s\n\tBODY: %s', method, str(headers), path, 'binary body')

        inittime = time.time()

        for idx in range(5):
            try:
                resp_txt = self._conn.rest_immediate(str1)
                break
            except Blob2OverrideError as excp:
                if idx == 4:
                    raise Blob2OverrideError(2)
                else:
                    continue

        endtime = time.time()

        LOGGER.info("iLO Response Time to %s: %s secs.", path, str(endtime-inittime))
        #Dummy response to support a bad host response
        if len(resp_txt) == 0:
            resp_txt = "HTTP/1.1 500 Not Found\r\nAllow: " \
            "GET\r\nCache-Control: no-cache\r\nContent-length: " \
            "0\r\nContent-type: text/html\r\nDate: Tues, 1 Apr 2025 " \
            "00:00:01 GMT\r\nServer: " \
            "HP-iLO-Server/1.30\r\nX_HP-CHRP-Service-Version: 1.0.3\r\n\r\n\r\n"

        restreq = RestRequest(path, method, data=body, url=self.base_url)
        rest_response = RisRestResponse(restreq, resp_txt)

        if rest_response.status in range(300, 399) and rest_response.status != 304:
            newloc = rest_response.getheader("location")
            newurl = urlparse(newloc)

            rest_response = self.rest_request(newurl.path, method, args, oribody, headers)

        try:
            if rest_response.getheader('content-encoding') == 'gzip':
                if hasattr(gzip, "decompress"):
                    rest_response.read = gzip.decompress(rest_response.ori)
                else:
                    compressedfile = BytesIO(rest_response.ori)
                    decompressedfile = gzip.GzipFile(fileobj=compressedfile)
                    rest_response.read = decompressedfile.read()
        except Exception:
            pass
        if LOGGER.isEnabledFor(logging.DEBUG):
            headerstr = ''
            headerget = rest_response.getheaders()
            for header in headerget:
                headerstr += '\t' + header + ': ' + headerget[header] + '\n'
            try:
                LOGGER.debug('Blobstore RESPONSE for %s:\nCode: %s\nHeaders:'\
                            '\n%s\nBody of %s: %s', rest_response.request.path,\
                            str(rest_response._http_response.status)+ ' ' + \
                            rest_response._http_response.reason, \
                            headerstr, rest_response.request.path, rest_response.read)
            except:
                LOGGER.debug('Blobstore RESPONSE for %s:\nCode:%s', \
                             rest_response.request.path, rest_response)
        return rest_response

    def cert_login(self):
        """Login using a certificate **NOT IMPLEMENTED**"""
        #TODO: Implement
        raise NotImplementedError()
